/* wptKeyserverDlg.cpp - Keyserver dialog
 *	Copyright (C) 2000-2006 Timo Schulz
 *	Copyright (C) 2005 g10 Code GmbH
 *
 * This file is part of WinPT.
 *
 * WinPT is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * WinPT is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with WinPT; if not, write to the Free Software Foundation, 
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <windows.h>
#include <commctrl.h>
#include <malloc.h>

#include "resource.h"
#include "wptKeyserver.h"
#include "wptErrors.h"
#include "wptTypes.h"
#include "wptCommonCtl.h"
#include "wptNLS.h"
#include "wptW32API.h"
#include "wptVersion.h"
#include "wptGPG.h"
#include "wptGPGME.h"
#include "wptContext.h" /* for passphrase_s */
#include "wptKeyManager.h"
#include "wptDlgs.h"
#include "wptUTF8.h"


char* get_reg_entry_keyserver (const char *);
int set_reg_entry_keyserver (const char *, const char *);

enum { KS_COL_NAME = 0, KS_COL_PORT };

/* Callback context to modify (add, edit) keyserver items. */
struct keyserver_rec_s {
    char name[200];	/* hostname */
    int  proto;		/* protocol */
    int  port;		/* used port */
    bool mode;		/* true = edit */
    bool cancel;
};
typedef struct keyserver_rec_s *keyserver_rec_t;

/* Print out keyserver error and a possible Winsock error. */
static void
hkp_err_box (HWND dlg, const char *host, WORD port, int rc)
{
    const char *err = kserver_strerror ();
    const char *ws_err = wsock_strerror ();
    
    if (ws_err && strlen (ws_err) > 0)
	log_box (_("Keyserver"), MB_ERR, "%s:%d: %s", host, port, ws_err);
    else
	log_box (_("Keyserver"), MB_ERR, "%s:%d: %s\n%s",
		 host, port, winpt_strerror (rc), err? err : "");
}


/* Send the extracted key with keyid @pattern to the 
   keyserver @kserver (port @port).
   Return value: 0 on success. */
int
hkp_send_key (HWND dlg, const char *kserver, WORD port, const char *pattern)
{
    GPGME *gpg;
    gpgme_error_t ec;
    char *rawkey = NULL;
    char msg[384];
    
    gpg = new GPGME ();
    gpg->setArmor (true);
    ec = gpg->exportToBuffer (pattern, &rawkey);
    if (ec) {
        msg_box (dlg, gpgme_strerror (ec), _("Export"), MB_ERR);
        goto leave;
    }
    op_begin ();
    ec = kserver_sendkey (kserver, port, rawkey, strlen (rawkey));
    op_end ();
    if (ec) {
        hkp_err_box (dlg, kserver, port, ec);
        goto leave;
    }
    
    _snprintf (msg, sizeof (msg) -1, _("Key '%s' successfully sent"), pattern);
    status_box (dlg, msg, _("GnuPG status"));
    
leave:
    delete gpg;
    free_if_alloc (rawkey);
    return ec? -1 : 0;
}



/* Show all received keys from the keyserver. */
static void
show_imported_keys (gpgme_import_result_t res)
{
    gpgme_import_status_t t;
    gpgme_key_t key;
    gpgme_data_t msg;
    gpgme_ctx_t ctx;
    gpgme_error_t err;
    const char *s;
    char *p, *uid;
    size_t n=0;

    if (!res)
	return;

    err = gpgme_data_new (&msg);
    if (err)
	BUG (NULL);
    err = gpgme_new (&ctx);
    if (err)
	BUG (NULL);
    if (res->considered > 1) {
	s = _("WARNING: multiple keys matched request.\n\n");
	gpgme_data_write (msg, s, strlen (s));
    }

    if (res->unchanged == res->considered)
	s = _("Key(s) successfully received but nothing was changed.");
    else
	s = _("Key(s) sucessfully received and imported.");
    gpgme_data_write (msg, s, strlen (s));
    gpgme_data_write (msg, "\n\n", 2);

    for (t=res->imports; t; t = t->next) {
	if (!gpgme_get_key (ctx, t->fpr, &key, 0)) {
	    s = uid = utf8_to_native (key->uids->uid);
	    gpgme_data_write (msg, s, strlen (s));
	    gpgme_data_write (msg, "\n", 1);
	    gpgme_key_release (key);
	    free (uid);
	}
    }
    gpgme_data_write (msg, "\0", 1);
    p = gpgme_data_release_and_get_mem (msg, &n);
    if (p != NULL) {
	msg_box (NULL, p, _("Imported Keys"), MB_INFO);
	gpgme_free (p);
    }
    gpgme_release (ctx);
}


/* Receive a key from the keyserver @kserver (port @port)
   with the pattern @pattern.
   Return value: 0 on success. */
static int
keyserver_recv_key (HWND dlg, const char *kserver, WORD port,
		    const char *pattern, int proto, int flags,
		    char **r_fpr)
{
    gpgme_ctx_t ctx;
    gpgme_data_t keydata;
    gpgme_import_result_t import_res = NULL;
    gpgme_error_t err;
    char *rawkey = NULL;
    int keylen = 0;
    int rc;

    if (proto == KSPROTO_LDAP) {
        rc = ldap_recvkey (kserver, pattern, &rawkey, &keylen);
        if (rc) {
	    msg_box (dlg, _("LDAP key import failed.\n"
			    "Please make sure you have an online connection"
			    " and gpgkeys_ldap.exe is installed"),
			    _("Keyserver"), MB_ERR);
	    free_if_alloc (rawkey);
            return rc;
	}
    }
    else if (proto == KSPROTO_FINGER) {
	rc = finger_recvkey (kserver, pattern, &rawkey, &keylen);
	if (rc) {
	    log_box (_("Keyserver"), MB_ERR, 
		     _("Finger key import failed: %s\n"), winpt_strerror (rc));
	    free_if_alloc (rawkey);
	    return rc;
	}
    }
    else if ((rc = kserver_recvkey (kserver, port, 
				    kserver_check_keyid (pattern),
				    &rawkey, &keylen))) {
        hkp_err_box (dlg, kserver, port, rc);
	free_if_alloc (rawkey);
        return rc;
    }

    if (!strstr (rawkey, "BEGIN PGP PUBLIC KEY BLOCK")) {
        msg_box (dlg, _("This is not a valid OpenPGP key."),
	    _("Keyserver"), MB_ERR);
	goto leave;
    }
    if (gpgme_new (&ctx))
	BUG (NULL);
    gpgme_data_new_from_mem (&keydata, rawkey, strlen (rawkey), 1);
    err = gpgme_op_import (ctx, keydata);
    if (err) {
        msg_box (dlg, gpgme_strerror (err), _("Import"), MB_ERR);
	goto leave;
    }
    import_res = gpgme_op_import_result (ctx);
    if (import_res && r_fpr)
	*r_fpr = m_strdup (import_res->imports->fpr);
    
    /* if we use the refresh mode, a lot of keys will be fetched and thus only
       a summarize at the end is presented and not for each key. */
    if (!(flags & KM_KS_REFRESH)) {
	show_imported_keys (import_res);
	if (import_res && import_res->unchanged == import_res->considered) {
	    rc = WPTERR_GENERAL; /* no keys updated. */
	    goto leave;
	}
    }

leave:
    free_if_alloc (rawkey);
    gpgme_release (ctx);
    gpgme_data_release (keydata);

    return rc;
}


/* Alias for keyserver_recv_key but without the ability to return 
   the fingerprint of the key. */
int
hkp_recv_key (HWND dlg, const char *kserver, WORD port,
	      const char *pattern, int proto, int flags)
{
    return keyserver_recv_key (dlg, kserver, port, pattern, 
			       proto, flags, NULL);
}

int
hkp_recv_key2 (HWND dlg, const char *kserver, WORD port,
	       const char *pattern, int proto, char **r_fpr)
{
    return keyserver_recv_key (dlg, kserver, port, pattern, proto, 0,
			       r_fpr);
}


#define my_iskeychar(a) (((a) >='0' && (a) <= '9' ) || ((a) >= 'A' && (a) <= 'F'))

static int
check_pattern (const char *pattern)
{
    int rc = WPTERR_GENERAL;
    
    /* do not allow white spaces. */
    if (strchr (pattern, ' '))
	return rc;

    if (strstr (pattern, "0x"))
	pattern += 2;

    if (((my_iskeychar (pattern[0])) && (strlen (pattern)  == 8)) ||
         (my_iskeychar (pattern[0])) && (strlen (pattern) == 16))
        rc = 0; /* assume long or short key ID. */
    else if (strchr (pattern, '@') && strlen (pattern) >= 3)
        rc = 0; /* assume email address. */

    return rc;
}


/* Return human readable name for the proxy protocol. */
static const char*
name_from_proto (int proto)
{
    const char *s;

    switch (proto) {
    case PROXY_PROTO_NONE:	s = ""; break;
    case PROXY_PROTO_HTTP:	s = "HTTP"; break;
    case PROXY_PROTO_SOCKS5:	s =  "SOCKS5"; break;
    default:			s= "HTTP"; break;
    }
    return s;
}


static int inline
kserver_get_pos (listview_ctrl_t lv)
{
    if (listview_count_items (lv, 0) == 1)
	return 0;
    return listview_get_curr_pos (lv);
}


static WORD inline
kserver_get_port (listview_ctrl_t lv)
{
    char buf[16];

    listview_get_item_text (lv, kserver_get_pos (lv), KS_COL_PORT, buf, 15);
    return (WORD)strtoul (buf, NULL, 10);
}


static void
set_default_ks (listview_ctrl_t lv)
{
    char *p, buf[200];
    int i;

    p = get_reg_entry_keyserver ("Default");
    if (!p)
	return;
    for (i=0; i < listview_count_items (lv, 0); i++) {
	listview_get_item_text (lv, i, KS_COL_NAME, buf, sizeof (buf)-1);	
	if (!strcmp (buf, p)) {	
	    listview_select_one (lv, i);
	    break;
	}
    }
    free_if_alloc (p);
}


/* Save the selected keyserver from @lv as the default server. */
static int
save_default_ks (listview_ctrl_t lv)
{
    char buf[192], port[32];
    int idx, i;   

    idx = listview_get_curr_pos (lv);
    if (idx == -1) {
	msg_box (NULL, _("Please select one of the servers."), _("Keyserver"), MB_ERR);
	return -1;
    }
    listview_get_item_text (lv, idx, KS_COL_NAME, buf, sizeof (buf)-1);
    if (strncmp (buf, "http", 4) && strncmp (buf, "hkp", 3)) {
	msg_box (NULL, _("Only HTTP keyserver can be used."), 
		 _("Keyserver"), MB_ERR);
	return -1;
    }
    set_reg_entry_keyserver ("Default", buf);
    i = kserver_get_port (lv);
    sprintf (port, "%d", i);
    set_reg_entry_keyserver ("Default_Port", port);
    keyserver_set_default (buf, (WORD)i);
    return 0;
}


int
keyserver_list_build (listview_ctrl_t *r_lv, HWND hwnd)
{
    struct listview_column_s keyserver[] = {
        {0, 180, (char *)_("DNS Name")},
	{1,  55, (char *)_("Port")},
        {0,   0, NULL}
    };
    HICON ico[1];
    listview_ctrl_t lv;
    char buf[32];
    int j;

    ico[0] = LoadIcon (glob_hinst, (LPCTSTR)IDI_COMPUTER);
    listview_new (&lv, hwnd);    
    for (j=0; keyserver[j].fieldname; j++)
        listview_add_column (lv, &keyserver[j]);
    listview_set_image_list (lv, 16, 16, ico, 1);
    for (j = 0; j<MAX_KEYSERVERS; j++) {
	if (!server[j].used)
	    continue;
	listview_add_item_image (lv, " ", 0);
	listview_add_sub_item (lv, 0, KS_COL_NAME, server[j].name);
	sprintf (buf, "%d", server[j].port);
	listview_add_sub_item (lv, 0, KS_COL_PORT, buf);
    }
    if (listview_count_items (lv, 0) == 0) {
	listview_add_item (lv, "");
	listview_add_sub_item (lv, 0, KS_COL_NAME, DEF_HKP_KEYSERVER);
    }
    listview_set_ext_style (lv);
    *r_lv = lv;
    return 0;
}


static void
keyserver_remove (listview_ctrl_t lv)
{
    char name[100];
    int pos = listview_get_curr_pos (lv);
    int i;

    listview_get_item_text (lv, pos, KS_COL_NAME, name, sizeof (name)-1);
    for (i=0; i < MAX_KEYSERVERS; i++) {
	if (server[i].name && strcmp (server[i].name, name) == 0)
	    server[i].used = 0;
    }
    listview_del_item (lv, pos);
}


static void
keyserver_edit (listview_ctrl_t lv, keyserver_rec_t ctx)
{
    char buf[16];
    bool fnd = false;
    int i;

    for (i=0; i < MAX_KEYSERVERS; i++) {
    	if (server[i].name && strcmp (server[i].name, ctx->name) == 0) {
	    fnd = true;
	    break;
	}
    }
    if (!fnd)
	return;
    server[i].port = ctx->port;

    i = listview_get_curr_pos (lv);
    sprintf (buf, "%d", ctx->port);
    listview_add_sub_item (lv, i, KS_COL_NAME, ctx->name);
    listview_add_sub_item (lv, i, KS_COL_PORT, buf);
}

static void
keyserver_add (listview_ctrl_t lv, keyserver_rec_t ctx)
{
    char buf[16];
    bool fnd = false;
    int i;

    for (i=0; i < MAX_KEYSERVERS; i++) {
	if (server[i].used == 0) {
	    fnd = true;
	    break;
	}
    }
    if (!fnd) {
	msg_box (NULL, _("No space for new keyserver entry"),
		 _("Keyserver"), MB_ERR);
	return;
    }
    free_if_alloc (server[i].name);
    server[i].name = m_strdup (ctx->name);
    server[i].port = ctx->port;
    server[i].proto = ctx->proto;
    server[i].used = 1;

    sprintf (buf, "%d", ctx->port);
    listview_add_item (lv, " ");
    listview_add_sub_item (lv, 0, KS_COL_NAME, ctx->name);
    listview_add_sub_item (lv, 0, KS_COL_PORT, buf);
}


static void
fill_keyserv_types (HWND dlg)
{
    HWND h = GetDlgItem (dlg, IDC_KSERVADD_TYPES);
    combox_add_string (h, _("HKP Keyserver"));
    combox_add_string (h, _("LDAP Keyserver"));
    combox_add_string (h, _("Finger Keyserver"));
    SendMessage (h, CB_SETCURSEL, 0, 0);
}


/* Dialog box procedure for modify keyservers. */
BOOL CALLBACK
keyserver_modify_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static keyserver_rec_t ctx;
    char tmp[128];

    switch (msg) {
    case WM_INITDIALOG:
	fill_keyserv_types (dlg);
	ctx = (keyserver_rec_t)lparam;
	if (ctx->mode) {
	    SendDlgItemMessage (dlg, IDC_KSERVADD_TYPES, CB_SETCURSEL, ctx->proto, 0);
	    SetDlgItemText (dlg, IDC_KSERVADD_HOST, ctx->name);
	    SetDlgItemInt (dlg, IDC_KSERVADD_PORT, ctx->port, FALSE);
	    EnableWindow (GetDlgItem (dlg, IDC_KSERVADD_HOST), FALSE);
	    EnableWindow (GetDlgItem (dlg, IDC_KSERVADD_TYPES), FALSE);
	}
	else
	    SetDlgItemInt (dlg, IDC_KSERVADD_PORT, 11371, FALSE);	
	SetWindowText (dlg, _("Edit Keyserver"));
	SetDlgItemText (dlg, IDOK, _("&Add"));
	SetDlgItemText (dlg, IDCANCEL, _("&Cancel"));
	SetDlgItemText (dlg, IDC_KSERVADD_TYPEINF, _("Type:"));
	SetDlgItemText (dlg, IDC_KSERVADD_PORTINF, _("Port:"));
	SetDlgItemText (dlg, IDC_KSERVADD_HOSTINF, _("Host name:"));
	SetForegroundWindow (dlg);
	break;

    case WM_COMMAND:
	switch (LOWORD (wparam)) {
	case IDOK:
	    if (!GetDlgItemText (dlg, IDC_KSERVADD_HOST, tmp, sizeof (tmp)-1)) {
		msg_box (dlg, _("Please enter a host name"), _("Keyserver"), MB_ERR);
		return FALSE;
	    }	    
	    ctx->port = GetDlgItemInt (dlg, IDC_KSERVADD_PORT, NULL, FALSE);
	    if (ctx->port > 65535) {
		msg_box (dlg, _("Invalid port, valid numbers are < 65535"), _("Keyserver"), MB_ERR);
		return FALSE;
	    }
	    ctx->proto = SendDlgItemMessage (dlg, IDC_KSERVADD_TYPES, CB_GETCURSEL, 0, 0);
	    ctx->cancel = false;
	    ctx->name[0] = '\0';
	    if (!strstr (tmp, "://")) {
		switch (ctx->proto) {
		case KSPROTO_HTTP: strcpy (ctx->name, "http://"); break;
		case KSPROTO_LDAP: strcpy (ctx->name, "ldap://"); break;
		case KSPROTO_FINGER:strcpy (ctx->name, "finger://"); break;
		}
	    }
	    strcat (ctx->name, tmp);
	    ctx->cancel = false;
	    EndDialog (dlg, TRUE);
	    break;

	case IDCANCEL:
	    ctx->cancel = true;
	    EndDialog (dlg, FALSE);
	    break;
	}
	break;
    }

    return FALSE;
}

/* Dialog box procedure to access keyservers. */
BOOL CALLBACK
keyserver_dlg_proc (HWND dlg, UINT msg, WPARAM wparam, LPARAM lparam)
{
    static listview_ctrl_t lv = NULL;
    static int lv_idx = 0;
    keyserver_ctx ksc;
    struct keyserver_rec_s edit;
    char kserver[128], pattern[128];
    char proto[16], *fpr = NULL;
    int rc, proto_nr = 0;
    
    switch (msg) {
    case WM_INITDIALOG:
        SetWindowText (dlg, _("Keyserver Access"));
	SetDlgItemText (dlg, IDC_KEYSERVER_RECV, _("&Receive"));
        SetDlgItemText (dlg, IDC_KEYSERVER_SEND, 
                        _("Send key (default is receiving)"));
        SetDlgItemText (dlg, IDC_KEYSERVER_INFO,
                        _("Please enter the key ID or email address you search for"));
        SetDlgItemText (dlg, IDC_KEYSERVER_INDEX, _("&Search"));
	SetDlgItemText (dlg, IDC_KEYSERVER_PROXSETT, _("C&hange proxy"));
	SetDlgItemText (dlg, IDC_KEYSERVER_DEFAULT, _("Set &default"));
	SetDlgItemText (dlg, IDCANCEL, _("&Close"));
	
        keyserver_list_build (&lv, GetDlgItem (dlg, IDC_KEYSERVER_LIST));
	set_default_ks (lv);
	center_window (dlg, NULL);
        SetForegroundWindow (dlg);
        return TRUE;
        
    case WM_NOTIFY:
        NMHDR *notify;
        notify = (NMHDR *)lparam;
	if (!notify)
	    break;
        if (notify->code == NM_CLICK
            && notify->idFrom == IDC_KEYSERVER_LIST)
            lv_idx = listview_get_curr_pos (lv);
	else if (notify->code == NM_RCLICK &&
		 notify->idFrom == IDC_KEYSERVER_LIST) {
	    POINT p;
	    GetCursorPos (&p);
	    HMENU hm, pop;
	    hm = LoadMenu (glob_hinst, (LPCTSTR)IDR_WINPT_KEYSERVER_CTX);
	    pop = GetSubMenu (hm, 0);
	    set_menu_text (pop, ID_KSERVCTX_ADD, _("&Add"));
	    set_menu_text (pop, ID_KSERVCTX_DEL, _("&Remove"));
	    set_menu_text (pop, ID_KSERVCTX_EDIT, _("&Edit"));
	    TrackPopupMenu (pop, 0, p.x, p.y, 0, dlg, NULL);
	    DestroyMenu (hm);
	    DestroyMenu (pop);
	}

        return TRUE;
        
    case WM_DESTROY:
        if (lv) {
            listview_release (lv);
            lv = NULL;
        }
        lv_idx = 0;
        return FALSE;
        
    case WM_SYSCOMMAND:
        if (LOWORD (wparam) == SC_CLOSE)
            EndDialog (dlg, TRUE);
        return FALSE;
        
    case WM_COMMAND:
        switch (LOWORD (wparam)) {
        case IDC_KEYSERVER_PROXSETT:
            dialog_box_param (glob_hinst, (LPCTSTR)IDD_WINPT_KEYSERVER_PROXY, 
			      dlg, keyserver_proxy_dlg_proc, 0,
                              _("Proxy Settings"), IDS_WINPT_KEYSERVER_PROXY);
            return TRUE;
            
        case IDC_KEYSERVER_INDEX:
            if (!lv_idx) {
                lv_idx = kserver_get_pos (lv);
                if (lv_idx == -1) {
                    msg_box (dlg, _("Please select one of the keyservers."), 
			     _("Keyserver"), MB_INFO);
                    return FALSE;
                }
            }
            listview_get_item_text (lv, lv_idx, KS_COL_NAME, 
				    proto, sizeof (proto)-1);
	    if (!strncmp (proto, "ldap", 4)) {
                msg_box (dlg, _("This is not implemented yet!"), 
			 _("Keyserver"), MB_ERR);
                return FALSE;
            }
            listview_get_item_text (lv, lv_idx, KS_COL_NAME, 
				    kserver, sizeof (kserver)-1);
            if (!GetDlgItemText (dlg, IDC_KEYSERVER_SEARCH, pattern, sizeof (pattern)-1)) {
                msg_box (dlg, _("Please enter the search pattern."), 
			 _("Keyserver"), MB_INFO);
                return FALSE;
            }
            ksc.name = kserver;
            ksc.pattern = pattern;
	    ksc.port = kserver_get_port (lv);
            DialogBoxParam (glob_hinst, (LPCSTR)IDD_WINPT_HKPSEARCH, dlg,
                            hkpsearch_dlg_proc, (LPARAM) &ksc);
            return TRUE;
            
        case IDC_KEYSERVER_RECV:
            memset (&kserver, 0, sizeof (kserver));
            if (!lv_idx) {
                lv_idx = kserver_get_pos (lv);
                if (lv_idx == -1) {
                    msg_box (dlg, _("Please select one of the keyservers."), 
			    _("Keyserver"), MB_INFO);
                    return FALSE;
                }
            }
            listview_get_item_text (lv, lv_idx, KS_COL_NAME, 
				    proto, sizeof (proto)-1);
	    proto_nr = KSPROTO_HTTP;
	    if (!strncmp  (proto, "ldap", 4))
		proto_nr = KSPROTO_LDAP;
	    else if (!strncmp (proto, "finger", 6))
		proto_nr = KSPROTO_FINGER;
            listview_get_item_text (lv, lv_idx, KS_COL_NAME, 
				    kserver, sizeof (kserver)-1);
            if (!GetDlgItemText(dlg, IDC_KEYSERVER_SEARCH,
				pattern, sizeof (pattern)-1)) {
                msg_box (dlg, _("Please enter the search pattern."),
			 _("Keyserver"), MB_INFO);
                return FALSE;
            }
            if (proto_nr == KSPROTO_LDAP && strchr (pattern, '@')) {
                msg_box (dlg, _("Only keyids are allowed."),
			 _("Keyserver"), MB_INFO);
                return FALSE;
            }
	    else if (proto_nr == KSPROTO_FINGER) {
		if (strchr (pattern, '@') || strchr (pattern, ' ')) {
		    msg_box (dlg, _("Only enter the name of the user."), 
			     _("Keyserver"), MB_INFO);
		    return FALSE;
		}
	    }
            else if (check_pattern (pattern)) {
                msg_box (dlg, _("Only email addresses or keyids are allowed."),
			_("Keyserver"), MB_INFO);
                return FALSE;
            }
	    rc = keyserver_recv_key (dlg, kserver, kserver_get_port (lv),
				     pattern, proto_nr, 0,
				     &fpr);
	    if (!rc && fpr != NULL) {
		keycache_update (0, fpr);
		free_if_alloc (fpr);
	    }
            return TRUE;

	case IDC_KEYSERVER_DEFAULT:
	    save_default_ks (lv);
	    break;

	case ID_KSERVCTX_ADD:
	    memset (&edit, 0, sizeof (edit));
	    edit.mode = false;
	    dialog_box_param (glob_hinst, (LPCTSTR)IDD_WINPT_KEYSERVER_ADD, dlg,
				keyserver_modify_dlg_proc, 
				(LPARAM)&edit, _("Edit Keyserver"), 0);
	    if (!edit.cancel)
		keyserver_add (lv, &edit);
	    return TRUE;

	case ID_KSERVCTX_EDIT:
	    memset (&edit, 0, sizeof (edit));
	    edit.mode = true;
	    listview_get_item_text (lv, lv_idx, KS_COL_NAME, 
				    edit.name, sizeof (edit.name)-1);
	    listview_get_item_text (lv, lv_idx, KS_COL_PORT, pattern, 32);
	    edit.port = atoi (pattern);
	    if (!strncmp (edit.name, "hkp", 3) || !strncmp (edit.name, "http", 4))
		edit.proto = KSPROTO_HTTP;
	    else if (!strncmp (edit.name, "ldap", 4))
		edit.proto = KSPROTO_LDAP;
	    else
		edit.proto = KSPROTO_FINGER;
	    dialog_box_param (glob_hinst, (LPCTSTR)IDD_WINPT_KEYSERVER_ADD, dlg,
				keyserver_modify_dlg_proc, 
				(LPARAM)&edit, _("Edit Keyserver"), 0);
	    if (!edit.cancel)
		keyserver_edit (lv, &edit);
	    break;

	case ID_KSERVCTX_DEL:
	    keyserver_remove (lv);
	    return TRUE;
            
        case IDCANCEL:
            EndDialog (dlg, FALSE);
            return FALSE;
        }
        break;
    }
    
    return FALSE;
}
